@@ -14,34 +14,36 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import typing as T
-import itertools
-import os
-import subprocess
-import shutil
-import sys
-import signal
-import shlex
-from io import StringIO
-from ast import literal_eval
+from concurrent.futures import ProcessPoolExecutor, CancelledError
 from enum import Enum
-import tempfile
+from io import StringIO
 from pathlib import Path, PurePath
+import argparse
+import functools
+import itertools
+import json
+import multiprocessing
+import os
+import re
+import shlex
+import shutil
+import signal
+import subprocess
+import sys
+import tempfile
+import time
+import typing as T
+import xml.etree.ElementTree as ET
+
 from mesonbuild import build
 from mesonbuild import environment
 from mesonbuild import compilers
 from mesonbuild import mesonlib
 from mesonbuild import mlog
 from mesonbuild import mtest
-from mesonbuild.mesonlib import MachineChoice, stringlistify, Popen_safe
+from mesonbuild.mesonlib import MachineChoice, Popen_safe
 from mesonbuild.coredata import backendlist
-import argparse
-import json
-import xml.etree.ElementTree as ET
-import time
-import multiprocessing
-from concurrent.futures import ProcessPoolExecutor, CancelledError
-import re
+
 from run_tests import get_fake_options, run_configure, get_meson_script
 from run_tests import get_backend_commands, get_backend_args_for_dir, Backend
 from run_tests import ensure_backend_detects_changes
@@ -63,25 +65,132 @@ class BuildStep(Enum):
     validate = 6
 
 
-class TestResult:
-    def __init__(self, msg, step, stdo, stde, mlog, cicmds, conftime=0, buildtime=0, testtime=0):
-        self.msg = msg
-        self.step = step
-        self.stdo = stdo
-        self.stde = stde
-        self.mlog = mlog
+class TestResult(BaseException):
+    def __init__(self, cicmds):
+        self.msg = ''  # empty msg indicates test success
+        self.stdo = ''
+        self.stde = ''
+        self.mlog = ''
         self.cicmds = cicmds
-        self.conftime = conftime
-        self.buildtime = buildtime
-        self.testtime = testtime
+        self.conftime = 0
+        self.buildtime = 0
+        self.testtime = 0
 
+    def add_step(self, step, stdo, stde, mlog='', time=0):
+        self.step = step
+        self.stdo += stdo
+        self.stde += stde
+        self.mlog += mlog
+        if step == BuildStep.configure:
+            self.conftime = time
+        elif step == BuildStep.build:
+            self.buildtime = time
+        elif step == BuildStep.test:
+            self.testtime = time
 
+    def fail(self, msg):
+        self.msg = msg
+
+class InstalledFile:
+    def __init__(self, raw: T.Dict[str, str]):
+        self.path = raw['file']
+        self.typ = raw['type']
+        self.platform = raw.get('platform', None)
+        self.language = raw.get('language', 'c')  # type: str
+
+        version = raw.get('version', '')  # type: str
+        if version:
+            self.version = version.split('.')  # type: T.List[str]
+        else:
+            # split on '' will return [''], we want an empty list though
+            self.version = []
+
+    def get_path(self, compiler: str, env: environment.Environment) -> T.Optional[Path]:
+        p = Path(self.path)
+        canonical_compiler = compiler
+        if ((compiler in ['clang-cl', 'intel-cl']) or
+                (env.machines.host.is_windows() and compiler in {'pgi', 'dmd', 'ldc'})):
+            canonical_compiler = 'msvc'
+
+        has_pdb = False
+        if self.language in {'c', 'cpp'}:
+            has_pdb = canonical_compiler == 'msvc'
+        elif self.language == 'd':
+            # dmd's optlink does not genearte pdb iles
+            has_pdb = env.coredata.compilers.host['d'].linker.id in {'link', 'lld-link'}
+
+        # Abort if the platform does not match
+        matches = {
+            'msvc': canonical_compiler == 'msvc',
+            'gcc': canonical_compiler != 'msvc',
+            'cygwin': env.machines.host.is_cygwin(),
+            '!cygwin': not env.machines.host.is_cygwin(),
+        }.get(self.platform or '', True)
+        if not matches:
+            return None
+
+        # Handle the different types
+        if self.typ == 'file':
+            return p
+        elif self.typ == 'shared_lib':
+            if env.machines.host.is_windows() or env.machines.host.is_cygwin():
+                # Windows only has foo.dll and foo-X.dll
+                if len(self.version) > 1:
+                    return None
+                if self.version:
+                    p = p.with_name('{}-{}'.format(p.name, self.version[0]))
+                return p.with_suffix('.dll')
+
+            p = p.with_name('lib{}'.format(p.name))
+            if env.machines.host.is_darwin():
+                # MacOS only has libfoo.dylib and libfoo.X.dylib
+                if len(self.version) > 1:
+                    return None
+
+                # pathlib.Path.with_suffix replaces, not appends
+                suffix = '.dylib'
+                if self.version:
+                    suffix = '.{}{}'.format(self.version[0], suffix)
+            else:
+                # pathlib.Path.with_suffix replaces, not appends
+                suffix = '.so'
+                if self.version:
+                    suffix = '{}.{}'.format(suffix, '.'.join(self.version))
+            return p.with_suffix(suffix)
+        elif self.typ == 'exe':
+            if env.machines.host.is_windows() or env.machines.host.is_cygwin():
+                return p.with_suffix('.exe')
+        elif self.typ == 'pdb':
+            if self.version:
+                p = p.with_name('{}-{}'.format(p.name, self.version[0]))
+            return p.with_suffix('.pdb') if has_pdb else None
+        elif self.typ == 'implib' or self.typ == 'implibempty':
+            if env.machines.host.is_windows() and canonical_compiler == 'msvc':
+                # only MSVC doesn't generate empty implibs
+                if self.typ == 'implibempty' and compiler == 'msvc':
+                    return None
+                return p.parent / (re.sub(r'^lib', '', p.name) + '.lib')
+            elif env.machines.host.is_windows() or env.machines.host.is_cygwin():
+                return p.with_suffix('.dll.a')
+            else:
+                return None
+        elif self.typ == 'expr':
+            return Path(platform_fix_name(p.as_posix(), canonical_compiler, env))
+        else:
+            raise RuntimeError('Invalid installed file type {}'.format(self.typ))
+
+        return p
+
+@functools.total_ordering
 class TestDef:
     def __init__(self, path: Path, name: T.Optional[str], args: T.List[str], skip: bool = False):
         self.path = path
         self.name = name
         self.args = args
         self.skip = skip
+        self.env = os.environ.copy()
+        self.installed_files = []  # type: T.List[InstalledFile]
+        self.do_not_set_opts = []      # type: T.List[str]
 
     def __repr__(self) -> str:
         return '<{}: {:<48} [{}: {}] -- {}>'.format(type(self).__name__, str(self.path), self.name, self.args, self.skip)
@@ -91,6 +200,14 @@ class TestDef:
             return '{}   ({})'.format(self.path.as_posix(), self.name)
         return self.path.as_posix()
 
+    def __lt__(self, other: T.Any) -> bool:
+        if isinstance(other, TestDef):
+            # None is not sortable, so replace it with an empty string
+            s_id = int(self.path.name.split(' ')[0])
+            o_id = int(other.path.name.split(' ')[0])
+            return (s_id, self.path, self.name or '') < (o_id, other.path, other.name or '')
+        return NotImplemented
+
 class AutoDeletedDir:
     def __init__(self, d):
         self.dir = d
@@ -109,11 +226,13 @@ class AutoDeletedDir:
 failing_logs = []
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
 under_ci = 'CI' in os.environ
-under_old_os_ci = 'OLD_OS_CI' in os.environ
+under_xenial_ci = under_ci and ('XENIAL' in os.environ)
+skip_scientific = under_ci and ('SKIP_SCIENTIFIC' in os.environ)
 do_debug = under_ci or print_debug
 no_meson_log_msg = 'No meson-log.txt found.'
 
 system_compiler = None
+compiler_id_map = {}  # type: T.Dict[str, str]
 
 class StopException(Exception):
     def __init__(self):
@@ -133,17 +252,8 @@ def setup_commands(optbackend):
     compile_commands, clean_commands, test_commands, install_commands, \
         uninstall_commands = get_backend_commands(backend, do_debug)
 
-def get_relative_files_list_from_dir(fromdir: Path) -> T.List[Path]:
-    return [file.relative_to(fromdir) for file in fromdir.rglob('*') if file.is_file()]
-
-def platform_fix_name(fname: str, compiler, env) -> str:
-    # canonicalize compiler
-    if (compiler in {'clang-cl', 'intel-cl'} or
-       (env.machines.host.is_windows() and compiler == 'pgi')):
-        canonical_compiler = 'msvc'
-    else:
-        canonical_compiler = compiler
-
+# TODO try to eliminate or at least reduce this function
+def platform_fix_name(fname: str, canonical_compiler: str, env: environment.Environment) -> str:
     if '?lib' in fname:
         if env.machines.host.is_windows() and canonical_compiler == 'msvc':
             fname = re.sub(r'lib/\?lib(.*)\.', r'bin/\1.', fname)
@@ -160,31 +270,6 @@ def platform_fix_name(fname: str, compiler, env) -> str:
         else:
             fname = re.sub(r'\?lib', 'lib', fname)
 
-    if fname.endswith('?exe'):
-        fname = fname[:-4]
-        if env.machines.host.is_windows() or env.machines.host.is_cygwin():
-            return fname + '.exe'
-
-    if fname.startswith('?msvc:'):
-        fname = fname[6:]
-        if canonical_compiler != 'msvc':
-            return None
-
-    if fname.startswith('?gcc:'):
-        fname = fname[5:]
-        if canonical_compiler == 'msvc':
-            return None
-
-    if fname.startswith('?cygwin:'):
-        fname = fname[8:]
-        if not env.machines.host.is_cygwin():
-            return None
-
-    if fname.startswith('?!cygwin:'):
-        fname = fname[9:]
-        if env.machines.host.is_cygwin():
-            return None
-
     if fname.endswith('?so'):
         if env.machines.host.is_windows() and canonical_compiler == 'msvc':
             fname = re.sub(r'lib/([^/]*)\?so$', r'bin/\1.dll', fname)
@@ -204,53 +289,28 @@ def platform_fix_name(fname: str, compiler, env) -> str:
         else:
             return fname[:-3] + '.so'
 
-    if fname.endswith('?implib') or fname.endswith('?implibempty'):
-        if env.machines.host.is_windows() and canonical_compiler == 'msvc':
-            # only MSVC doesn't generate empty implibs
-            if fname.endswith('?implibempty') and compiler == 'msvc':
-                return None
-            return re.sub(r'/(?:lib|)([^/]*?)\?implib(?:empty|)$', r'/\1.lib', fname)
-        elif env.machines.host.is_windows() or env.machines.host.is_cygwin():
-            return re.sub(r'\?implib(?:empty|)$', r'.dll.a', fname)
-        else:
-            return None
-
     return fname
 
-def validate_install(srcdir: str, installdir: Path, compiler, env) -> str:
-    # List of installed files
-    info_file = Path(srcdir) / 'installed_files.txt'
-    installdir = Path(installdir)
-    # If this exists, the test does not install any other files
-    noinst_file = Path('usr/no-installed-files')
-    expected = {}  # type: T.Dict[Path, bool]
+def validate_install(test: TestDef, installdir: Path, compiler: str, env: environment.Environment) -> str:
+    expected_raw = [x.get_path(compiler, env) for x in test.installed_files]
+    expected = {Path(x): False for x in expected_raw if x}
+    found = [x.relative_to(installdir) for x in installdir.rglob('*') if x.is_file() or x.is_symlink()]
     ret_msg = ''
-    # Generate list of expected files
-    if (installdir / noinst_file).is_file():
-        expected[noinst_file] = False
-    elif info_file.is_file():
-        with info_file.open() as f:
-            for line in f:
-                line = platform_fix_name(line.strip(), compiler, env)
-                if line:
-                    expected[Path(line)] = False
-    # Check if expected files were found
-    for fname in expected:
-        file_path = installdir / fname
-        if file_path.is_file() or file_path.is_symlink():
-            expected[fname] = True
-    for (fname, found) in expected.items():
-        if not found:
-            ret_msg += 'Expected file {} missing.\n'.format(fname)
-    # Check if there are any unexpected files
-    found = get_relative_files_list_from_dir(installdir)
+    # Mark all found files as found and detect unexpected files
     for fname in found:
         if fname not in expected:
             ret_msg += 'Extra file {} found.\n'.format(fname)
+            continue
+        expected[fname] = True
+    # Check if expected files were found
+    for p, f in expected.items():
+        if not f:
+            ret_msg += 'Expected file {} missing.\n'.format(p)
+    # List dir content on error
     if ret_msg != '':
         ret_msg += '\nInstall dir contents:\n'
         for i in found:
-            ret_msg += '  - {}'.format(i)
+            ret_msg += '  - {}\n'.format(i)
     return ret_msg
 
 def log_text_file(logfile, testdir, stdo, stde):
@@ -321,6 +381,18 @@ def run_ci_commands(raw_log: str) -> T.List[str]:
     return res
 
 
+# There are some class variables and such that cahce
+# information. Clear all of these. The better solution
+# would be to change the code so that no state is persisted
+# but that would be a lot of work given that Meson was originally
+# coded to run as a batch process.
+def clear_internal_caches():
+    import mesonbuild.interpreterbase
+    from mesonbuild.dependencies import CMakeDependency
+    from mesonbuild.mesonlib import PerMachine
+    mesonbuild.interpreterbase.FeatureNew.feature_registry = {}
+    CMakeDependency.class_cmakeinfo = PerMachine(None, None)
+
 def run_test_inprocess(testdir):
     old_stdout = sys.stdout
     sys.stdout = mystdout = StringIO()
@@ -342,170 +414,184 @@ def run_test_inprocess(testdir):
         os.chdir(old_cwd)
     return max(returncode_test, returncode_benchmark), mystdout.getvalue(), mystderr.getvalue(), test_log
 
-def parse_test_args(testdir):
-    args = []
-    try:
-        with open(os.path.join(testdir, 'test_args.txt'), 'r') as f:
-            content = f.read()
-            try:
-                args = literal_eval(content)
-            except Exception:
-                raise Exception('Malformed test_args file.')
-            args = stringlistify(args)
-    except FileNotFoundError:
-        pass
-    return args
-
 # Build directory name must be the same so Ccache works over
 # consecutive invocations.
-def create_deterministic_builddir(src_dir, name):
+def create_deterministic_builddir(test: TestDef) -> str:
     import hashlib
-    if name:
-        src_dir += name
+    src_dir = test.path.as_posix()
+    if test.name:
+        src_dir += test.name
     rel_dirname = 'b ' + hashlib.sha256(src_dir.encode(errors='ignore')).hexdigest()[0:10]
     os.mkdir(rel_dirname)
     abs_pathname = os.path.join(os.getcwd(), rel_dirname)
     return abs_pathname
 
-def run_test(skipped, testdir, name, extra_args, compiler, backend, flags, commands, should_fail):
-    if skipped:
+def run_test(test: TestDef, extra_args, compiler, backend, flags, commands, should_fail):
+    if test.skip:
         return None
-    with AutoDeletedDir(create_deterministic_builddir(testdir, name)) as build_dir:
+    with AutoDeletedDir(create_deterministic_builddir(test)) as build_dir:
         with AutoDeletedDir(tempfile.mkdtemp(prefix='i ', dir=os.getcwd())) as install_dir:
             try:
-                return _run_test(testdir, build_dir, install_dir, extra_args, compiler, backend, flags, commands, should_fail)
+                return _run_test(test, build_dir, install_dir, extra_args, compiler, backend, flags, commands, should_fail)
+            except TestResult as r:
+                return r
             finally:
                 mlog.shutdown() # Close the log file because otherwise Windows wets itself.
 
-def pass_prefix_to_test(dirname):
-    if '39 prefix absolute' in dirname:
-        return False
-    return True
-
-def pass_libdir_to_test(dirname):
-    if '8 install' in dirname:
-        return False
-    if '38 libdir must be inside prefix' in dirname:
-        return False
-    if '195 install_mode' in dirname:
-        return False
-    return True
-
-def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backend, flags, commands, should_fail):
+def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args, compiler, backend, flags, commands, should_fail):
     compile_commands, clean_commands, install_commands, uninstall_commands = commands
-    test_args = parse_test_args(testdir)
     gen_start = time.time()
-    setup_env = None
     # Configure in-process
-    if pass_prefix_to_test(testdir):
-        gen_args = ['--prefix', '/usr']
-    else:
-        gen_args = []
-    if pass_libdir_to_test(testdir):
+    gen_args = []  # type: T.List[str]
+    if 'prefix' not in test.do_not_set_opts:
+        gen_args += ['--prefix', 'x:/usr'] if mesonlib.is_windows() else ['--prefix', '/usr']
+    if 'libdir' not in test.do_not_set_opts:
         gen_args += ['--libdir', 'lib']
-    gen_args += [testdir, test_build_dir] + flags + test_args + extra_args
-    nativefile = os.path.join(testdir, 'nativefile.ini')
-    if os.path.exists(nativefile):
-        gen_args.extend(['--native-file', nativefile])
-    crossfile = os.path.join(testdir, 'crossfile.ini')
-    if os.path.exists(crossfile):
-        gen_args.extend(['--cross-file', crossfile])
-    setup_env_file = os.path.join(testdir, 'setup_env.json')
-    if os.path.exists(setup_env_file):
-        setup_env = os.environ.copy()
-        with open(setup_env_file, 'r') as fp:
-            data = json.load(fp)
-            for key, val in data.items():
-                val = val.replace('@ROOT@', os.path.abspath(testdir))
-                setup_env[key] = val
-    (returncode, stdo, stde) = run_configure(gen_args, env=setup_env)
+    gen_args += [test.path.as_posix(), test_build_dir] + flags + extra_args
+    nativefile = test.path / 'nativefile.ini'
+    crossfile = test.path / 'crossfile.ini'
+    if nativefile.exists():
+        gen_args.extend(['--native-file', nativefile.as_posix()])
+    if crossfile.exists():
+        gen_args.extend(['--cross-file', crossfile.as_posix()])
+    (returncode, stdo, stde) = run_configure(gen_args, env=test.env)
     try:
         logfile = Path(test_build_dir, 'meson-logs', 'meson-log.txt')
         mesonlog = logfile.open(errors='ignore', encoding='utf-8').read()
     except Exception:
         mesonlog = no_meson_log_msg
     cicmds = run_ci_commands(mesonlog)
-    gen_time = time.time() - gen_start
+    testresult = TestResult(cicmds)
+    testresult.add_step(BuildStep.configure, stdo, stde, mesonlog, time.time() - gen_start)
     if should_fail == 'meson':
         if returncode == 1:
-            return TestResult('', BuildStep.configure, stdo, stde, mesonlog, cicmds, gen_time)
+            return testresult
         elif returncode != 0:
-            return TestResult('Test exited with unexpected status {}'.format(returncode), BuildStep.configure, stdo, stde, mesonlog, cicmds, gen_time)
+            testresult.fail('Test exited with unexpected status {}.'.format(returncode))
+            return testresult
         else:
-            return TestResult('Test that should have failed succeeded', BuildStep.configure, stdo, stde, mesonlog, cicmds, gen_time)
+            testresult.fail('Test that should have failed succeeded.')
+            return testresult
     if returncode != 0:
-        return TestResult('Generating the build system failed.', BuildStep.configure, stdo, stde, mesonlog, cicmds, gen_time)
+        testresult.fail('Generating the build system failed.')
+        return testresult
     builddata = build.load(test_build_dir)
-    # Touch the meson.build file to force a regenerate so we can test that
-    # regeneration works before a build is run.
-    ensure_backend_detects_changes(backend)
-    os.utime(os.path.join(testdir, 'meson.build'))
-    # Build with subprocess
     dir_args = get_backend_args_for_dir(backend, test_build_dir)
-    build_start = time.time()
-    pc, o, e = Popen_safe(compile_commands + dir_args, cwd=test_build_dir)
-    build_time = time.time() - build_start
-    stdo += o
-    stde += e
-    if should_fail == 'build':
+
+    # Build with subprocess
+    def build_step():
+        build_start = time.time()
+        pc, o, e = Popen_safe(compile_commands + dir_args, cwd=test_build_dir)
+        testresult.add_step(BuildStep.build, o, e, '', time.time() - build_start)
+        if should_fail == 'build':
+            if pc.returncode != 0:
+                raise testresult
+            testresult.fail('Test that should have failed to build succeeded.')
+            raise testresult
         if pc.returncode != 0:
-            return TestResult('', BuildStep.build, stdo, stde, mesonlog, cicmds, gen_time)
-        return TestResult('Test that should have failed to build succeeded', BuildStep.build, stdo, stde, mesonlog, cicmds, gen_time)
-    if pc.returncode != 0:
-        return TestResult('Compiling source code failed.', BuildStep.build, stdo, stde, mesonlog, cicmds, gen_time, build_time)
-    # Touch the meson.build file to force a regenerate so we can test that
-    # regeneration works after a build is complete.
-    ensure_backend_detects_changes(backend)
-    os.utime(os.path.join(testdir, 'meson.build'))
-    test_start = time.time()
+            testresult.fail('Compiling source code failed.')
+            raise testresult
+
+    # Touch the meson.build file to force a regenerate
+    def force_regenerate():
+        ensure_backend_detects_changes(backend)
+        os.utime(str(test.path / 'meson.build'))
+
+    # just test building
+    build_step()
+
+    # test that regeneration works for build step
+    force_regenerate()
+    build_step()  # TBD: assert nothing gets built after the regenerate?
+
+    # test that regeneration works for test step
+    force_regenerate()
+
     # Test in-process
+    clear_internal_caches()
+    test_start = time.time()
     (returncode, tstdo, tstde, test_log) = run_test_inprocess(test_build_dir)
-    test_time = time.time() - test_start
-    stdo += tstdo
-    stde += tstde
-    mesonlog += test_log
+    testresult.add_step(BuildStep.test, tstdo, tstde, test_log, time.time() - test_start)
     if should_fail == 'test':
         if returncode != 0:
-            return TestResult('', BuildStep.test, stdo, stde, mesonlog, cicmds, gen_time)
-        return TestResult('Test that should have failed to run unit tests succeeded', BuildStep.test, stdo, stde, mesonlog, cicmds, gen_time)
+            return testresult
+        testresult.fail('Test that should have failed to run unit tests succeeded.')
+        return testresult
     if returncode != 0:
-        return TestResult('Running unit tests failed.', BuildStep.test, stdo, stde, mesonlog, cicmds, gen_time, build_time, test_time)
+        testresult.fail('Running unit tests failed.')
+        return testresult
+
     # Do installation, if the backend supports it
     if install_commands:
         env = os.environ.copy()
         env['DESTDIR'] = install_dir
         # Install with subprocess
         pi, o, e = Popen_safe(install_commands, cwd=test_build_dir, env=env)
-        stdo += o
-        stde += e
+        testresult.add_step(BuildStep.install, o, e)
         if pi.returncode != 0:
-            return TestResult('Running install failed.', BuildStep.install, stdo, stde, mesonlog, cicmds, gen_time, build_time, test_time)
+            testresult.fail('Running install failed.')
+            return testresult
+
     # Clean with subprocess
     env = os.environ.copy()
     pi, o, e = Popen_safe(clean_commands + dir_args, cwd=test_build_dir, env=env)
-    stdo += o
-    stde += e
+    testresult.add_step(BuildStep.clean, o, e)
     if pi.returncode != 0:
-        return TestResult('Running clean failed.', BuildStep.clean, stdo, stde, mesonlog, cicmds, gen_time, build_time, test_time)
+        testresult.fail('Running clean failed.')
+        return testresult
+
+    # Validate installed files
+    testresult.add_step(BuildStep.install, '', '')
     if not install_commands:
-        return TestResult('', BuildStep.install, '', '', mesonlog, cicmds, gen_time, build_time, test_time)
-    return TestResult(validate_install(testdir, install_dir, compiler, builddata.environment),
-                      BuildStep.validate, stdo, stde, mesonlog, cicmds, gen_time, build_time, test_time)
+        return testresult
+    install_msg = validate_install(test, Path(install_dir), compiler, builddata.environment)
+    if install_msg:
+        testresult.fail('\n' + install_msg)
+        return testresult
+
+    return testresult
 
 def gather_tests(testdir: Path) -> T.List[TestDef]:
-    tests = [t.name for t in testdir.glob('*') if t.is_dir()]
+    tests = [t.name for t in testdir.iterdir() if t.is_dir()]
     tests = [t for t in tests if not t.startswith('.')]  # Filter non-tests files (dot files, etc)
-    tests = [TestDef(testdir / t, None, []) for t in tests]
-    all_tests = []
-    for t in tests:
-        matrix_file = t.path / 'test_matrix.json'
-        if not matrix_file.is_file():
+    test_defs = [TestDef(testdir / t, None, []) for t in tests]
+    all_tests = []  # type: T.List[TestDef]
+    for t in test_defs:
+        test_def_file = t.path / 'test.json'
+        if not test_def_file.is_file():
             all_tests += [t]
             continue
 
-        # Build multiple tests from matrix definition
+        test_def = json.loads(test_def_file.read_text())
+
+        # Handle additional environment variables
+        env = {}  # type: T.Dict[str, str]
+        if 'env' in test_def:
+            assert isinstance(test_def['env'], dict)
+            env = test_def['env']
+            for key, val in env.items():
+                val = val.replace('@ROOT@', t.path.resolve().as_posix())
+                env[key] = val
+
+        # Handle installed files
+        installed = []  # type: T.List[InstalledFile]
+        if 'installed' in test_def:
+            installed = [InstalledFile(x) for x in test_def['installed']]
+
+        # Handle the do_not_set_opts list
+        do_not_set_opts = test_def.get('do_not_set_opts', [])  # type: T.List[str]
+
+        # Skip the matrix code and just update the existing test
+        if 'matrix' not in test_def:
+            t.env.update(env)
+            t.installed_files = installed
+            t.do_not_set_opts = do_not_set_opts
+            all_tests += [t]
+            continue
+
+        # 'matrix; entry is present, so build multiple tests from matrix definition
         opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
-        matrix = json.loads(matrix_file.read_text())
+        matrix = test_def['matrix']
         assert "options" in matrix
         for key, val in matrix["options"].items():
             assert isinstance(val, list)
@@ -515,12 +601,24 @@ def gather_tests(testdir: Path) -> T.List[TestDef]:
                 assert "val" in i
                 skip = False
 
+                # Add an empty matrix entry
+                if i['val'] is None:
+                    tmp_opts += [(None, False)]
+                    continue
+
                 # Skip the matrix entry if environment variable is present
                 if 'skip_on_env' in i:
-                    for env in i['skip_on_env']:
-                        if env in os.environ:
+                    for skip_env_var in i['skip_on_env']:
+                        if skip_env_var in os.environ:
                             skip = True
 
+                # Only run the test if all compiler ID's match
+                if 'compilers' in i:
+                    for lang, id_list in i['compilers'].items():
+                        if lang not in compiler_id_map or compiler_id_map[lang] not in id_list:
+                            skip = True
+                            break
+
                 tmp_opts += [('{}={}'.format(key, i['val']), skip)]
 
             if opt_list:
@@ -532,16 +630,35 @@ def gather_tests(testdir: Path) -> T.List[TestDef]:
             else:
                 opt_list = [[x] for x in tmp_opts]
 
-        for i in opt_list:
-            name = ' '.join([x[0] for x in i])
-            opts = ['-D' + x[0] for x in i]
-            skip = any([x[1] for x in i])
-            all_tests += [TestDef(t.path, name, opts, skip)]
+        # Exclude specific configurations
+        if 'exclude' in matrix:
+            assert isinstance(matrix['exclude'], list)
+            new_opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
+            for i in opt_list:
+                exclude = False
+                opt_names = [x[0] for x in i]
+                for j in matrix['exclude']:
+                    ex_list = ['{}={}'.format(k, v) for k, v in j.items()]
+                    if all([x in opt_names for x in ex_list]):
+                        exclude = True
+                        break
 
-    all_tests = [(int(t.path.name.split()[0]), t.name or '', t) for t in all_tests]
-    all_tests.sort()
-    all_tests = [t[2] for t in all_tests]
-    return all_tests
+                if not exclude:
+                    new_opt_list += [i]
+
+            opt_list = new_opt_list
+
+        for i in opt_list:
+            name = ' '.join([x[0] for x in i if x[0] is not None])
+            opts = ['-D' + x[0] for x in i if x[0] is not None]
+            skip = any([x[1] for x in i])
+            test = TestDef(t.path, name, opts, skip)
+            test.env.update(env)
+            test.installed_files = installed
+            test.do_not_set_opts = do_not_set_opts
+            all_tests += [test]
+
+    return sorted(all_tests)
 
 def have_d_compiler():
     if shutil.which("ldc2"):
@@ -601,12 +718,16 @@ def have_java():
 
 def skippable(suite, test):
     # Everything is optional when not running on CI, or on Ubuntu 16.04 CI
-    if not under_ci or under_old_os_ci:
+    if not under_ci or under_xenial_ci:
         return True
 
     if not suite.endswith('frameworks'):
         return True
 
+    # this test assumptions aren't valid for Windows paths
+    if test.endswith('38 libdir must be inside prefix'):
+        return True
+
     # gtk-doc test may be skipped, pending upstream fixes for spaces in
     # filenames landing in the distro used for CI
     if test.endswith('10 gtk-doc'):
@@ -624,6 +745,11 @@ def skippable(suite, test):
     if test.endswith('29 blocks'):
         return True
 
+    # Scientific libraries are skippable on certain systems
+    # See the discussion here: https://github.com/mesonbuild/meson/pull/6562
+    if any([test.endswith(x) for x in ['17 mpi', '25 hdf5', '30 scalapack']]) and skip_scientific:
+        return True
+
     # No frameworks test should be skipped on linux CI, as we expect all
     # prerequisites to be installed
     if mesonlib.is_linux():
@@ -814,7 +940,8 @@ def _run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
                 suite_args = ['--fatal-meson-warnings']
                 should_fail = name.split('warning-')[1]
 
-            result = executor.submit(run_test, skipped or t.skip, t.path.as_posix(), t.name, extra_args + suite_args + t.args,
+            t.skip = skipped or t.skip
+            result = executor.submit(run_test, t, extra_args + suite_args + t.args,
                                      system_compiler, backend, backend_flags, commands, should_fail)
             futures.append((testname, t, result))
         for (testname, t, result) in futures:
@@ -918,6 +1045,8 @@ def check_format():
             continue
         if 'meson-logs' in root or 'meson-private' in root:
             continue
+        if '__CMake_build' in root:
+            continue
         if '.eggs' in root or '_cache' in root:  # e.g. .mypy_cache
             continue
         for fname in filenames:
@@ -927,13 +1056,13 @@ def check_format():
                     continue
                 check_file(root / file)
 
-def check_meson_commands_work():
+def check_meson_commands_work(options):
     global backend, compile_commands, test_commands, install_commands
     testdir = PurePath('test cases', 'common', '1 trivial').as_posix()
     meson_commands = mesonlib.python_command + [get_meson_script()]
     with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
         print('Checking that configuring works...')
-        gen_cmd = meson_commands + [testdir, build_dir] + backend_flags
+        gen_cmd = meson_commands + [testdir, build_dir] + backend_flags + options.extra_args
         pc, o, e = Popen_safe(gen_cmd)
         if pc.returncode != 0:
             raise RuntimeError('Failed to configure {!r}:\n{}\n{}'.format(testdir, e, o))
@@ -953,19 +1082,23 @@ def check_meson_commands_work():
                 raise RuntimeError('Failed to install {!r}:\n{}\n{}'.format(testdir, e, o))
 
 
-def detect_system_compiler():
-    global system_compiler
+def detect_system_compiler(options):
+    global system_compiler, compiler_id_map
 
     with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
-        env = environment.Environment(None, build_dir, get_fake_options('/'))
+        fake_opts = get_fake_options('/')
+        if options.cross_file:
+            fake_opts.cross_file = [options.cross_file]
+        env = environment.Environment(None, build_dir, fake_opts)
         print()
         for lang in sorted(compilers.all_languages):
             try:
                 comp = env.compiler_from_language(lang, MachineChoice.HOST)
-                details = '%s %s' % (' '.join(comp.get_exelist()), comp.get_version_string())
+                details = '{:<10} {} {}'.format('[' + comp.get_id() + ']', ' '.join(comp.get_exelist()), comp.get_version_string())
+                compiler_id_map[lang] = comp.get_id()
             except mesonlib.MesonException:
                 comp = None
-                details = 'not found'
+                details = '[not found]'
             print('%-7s: %s' % (lang, details))
 
             # note C compiler for later use by platform_fix_name()
@@ -981,7 +1114,7 @@ def print_tool_versions():
         {
             'tool': 'cmake',
             'args': ['--version'],
-            'regex': re.compile(r'^cmake version ([0-9]+(\.[0-9]+)*)$'),
+            'regex': re.compile(r'^cmake version ([0-9]+(\.[0-9]+)*(-[a-z0-9]+)?)$'),
             'match_group': 1,
         },
     ]
@@ -1018,16 +1151,19 @@ if __name__ == '__main__':
     parser.add_argument('--no-unittests', action='store_true',
                         help='Not used, only here to simplify run_tests.py')
     parser.add_argument('--only', help='name of test(s) to run', nargs='+', choices=ALL_TESTS)
+    parser.add_argument('--cross-file', action='store', help='File describing cross compilation environment.')
     options = parser.parse_args()
-    setup_commands(options.backend)
+    if options.cross_file:
+        options.extra_args += ['--cross-file', options.cross_file]
 
-    detect_system_compiler()
+    setup_commands(options.backend)
+    detect_system_compiler(options)
     print_tool_versions()
     script_dir = os.path.split(__file__)[0]
     if script_dir != '':
         os.chdir(script_dir)
     check_format()
-    check_meson_commands_work()
+    check_meson_commands_work(options)
     try:
         all_tests = detect_tests_to_run(options.only)
         (passing_tests, failing_tests, skipped_tests) = run_tests(all_tests, 'meson-test-run', options.failfast, options.extra_args)
