@@ -36,6 +36,24 @@ from mesonbuild.environment import Environment, detect_ninja
 from mesonbuild.coredata import backendlist
 
 NINJA_1_9_OR_NEWER = False
+NINJA_CMD = None
+# If we're on CI, just assume we have ninja in PATH and it's new enough because
+# we provide that. This avoids having to detect ninja for every subprocess unit
+# test that we run.
+if 'CI' in os.environ:
+    NINJA_1_9_OR_NEWER = True
+    NINJA_CMD = 'ninja'
+else:
+    # Look for 1.9 to see if https://github.com/ninja-build/ninja/issues/1219
+    # is fixed
+    NINJA_CMD = detect_ninja('1.9')
+    if NINJA_CMD is not None:
+        NINJA_1_9_OR_NEWER = True
+    else:
+        mlog.warning('Found ninja <1.9, tests will run slower', once=True)
+        NINJA_CMD = detect_ninja()
+if NINJA_CMD is None:
+    raise RuntimeError('Could not find Ninja v1.7 or newer')
 
 def guess_backend(backend, msbuild_exe: str):
     # Auto-detect backend if unspecified
@@ -202,22 +220,8 @@ def get_backend_commands(backend, debug=False):
         clean_cmd = cmd + ['-alltargets', 'clean', '-UseNewBuildSystem=FALSE']
         test_cmd = cmd + ['-target', 'RUN_TESTS']
     elif backend is Backend.ninja:
-        global NINJA_1_9_OR_NEWER
-        # Look for 1.9 to see if https://github.com/ninja-build/ninja/issues/1219
-        # is fixed, else require 1.6 for -w dupbuild=err
-        for v in ('1.9', '1.6'):
-            ninja_cmd = detect_ninja(v)
-            if ninja_cmd is not None:
-                if v == '1.9':
-                    NINJA_1_9_OR_NEWER = True
-                else:
-                    mlog.warning('Found ninja <1.9, tests will run slower', once=True)
-                    if 'CI' in os.environ and 'OLD_OS_CI' not in os.environ:
-                        raise RuntimeError('Require ninja >= 1.9 when running on Meson CI')
-                break
-        cmd = [ninja_cmd, '-w', 'dupbuild=err', '-d', 'explain']
-        if cmd[0] is None:
-            raise RuntimeError('Could not find Ninja v1.6 or newer')
+        global NINJA_CMD
+        cmd = [NINJA_CMD, '-w', 'dupbuild=err', '-d', 'explain']
         if debug:
             cmd += ['-v']
         clean_cmd = cmd + ['clean']
@@ -306,6 +310,7 @@ def print_system_info():
     print('Processor:', platform.processor())
     print('System:', platform.system())
     print('')
+    print(flush=True)
 
 def main():
     print_system_info()
@@ -313,7 +318,7 @@ def main():
     parser.add_argument('--cov', action='store_true')
     parser.add_argument('--backend', default=None, dest='backend',
                         choices=backendlist)
-    parser.add_argument('--cross', default=False, dest='cross', action='store_true')
+    parser.add_argument('--cross', default=[], dest='cross', action='append')
     parser.add_argument('--failfast', action='store_true')
     parser.add_argument('--no-unittests', action='store_true', default=False)
     (options, _) = parser.parse_known_args()
@@ -346,8 +351,6 @@ def main():
     if 'APPVEYOR' in os.environ and os.environ['arch'] == 'x86':
         os.environ.pop('platform')
     # Run tests
-    print(mlog.bold('Running unittests.').get_text(mlog.colorize_console))
-    print(flush=True)
     # Can't pass arguments to unit tests, so set the backend to use in the environment
     env = os.environ.copy()
     env['MESON_UNIT_TEST_BACKEND'] = backend.name
@@ -371,8 +374,11 @@ def main():
                 return returncode
             if no_unittests:
                 print('Skipping all unit tests.')
+                print(flush=True)
                 returncode = 0
             else:
+                print(mlog.bold('Running unittests.').get_text(mlog.colorize_console))
+                print(flush=True)
                 cmd = mesonlib.python_command + ['run_unittests.py', '-v']
                 if options.failfast:
                     cmd += ['--failfast']
@@ -383,21 +389,15 @@ def main():
             returncode += subprocess.call(cmd, env=env)
         else:
             cross_test_args = mesonlib.python_command + ['run_cross_test.py']
-            print(mlog.bold('Running armhf cross tests.').get_text(mlog.colorize_console))
-            print(flush=True)
-            cmd = cross_test_args + ['cross/ubuntu-armhf.txt']
-            if options.failfast:
-                cmd += ['--failfast']
-            returncode += subprocess.call(cmd, env=env)
-            if options.failfast and returncode != 0:
-                return returncode
-            print(mlog.bold('Running mingw-w64 64-bit cross tests.')
-                  .get_text(mlog.colorize_console))
-            print(flush=True)
-            cmd = cross_test_args + ['cross/linux-mingw-w64-64bit.txt']
-            if options.failfast:
-                cmd += ['--failfast']
-            returncode += subprocess.call(cmd, env=env)
+            for cf in options.cross:
+                print(mlog.bold('Running {} cross tests.'.format(cf)).get_text(mlog.colorize_console))
+                print(flush=True)
+                cmd = cross_test_args + ['cross/' + cf]
+                if options.failfast:
+                    cmd += ['--failfast']
+                returncode += subprocess.call(cmd, env=env)
+                if options.failfast and returncode != 0:
+                    return returncode
     return returncode
 
 if __name__ == '__main__':
